from youtube_autonomous.elements.validator.element_parameter_validator import ElementParameterValidator, ParameterValidator
from youtube_autonomous.segments.enums import SegmentType, EnhancementType
from youtube_autonomous.segments.builder.ai import create_ai_narration
from yta_general_utils.programming.parameter_obtainer import ParameterObtainer
from yta_general_utils.programming.parameter_validator import PythonValidator
from yta_general_utils.file.enums import FileTypeX
from yta_general_utils.programming.output import Output
from typing import Union


class ElementBuilder:
    # TODO: Maybe __init__ that returns the corresponding subclass
    # according to the provided 'type' (?)
    # https://stackoverflow.com/questions/9187388/possible-to-prevent-init-from-being-called

    @staticmethod
    def get_subclasses():
        from youtube_autonomous.elements.builder.ai_image_element_builder import AIImageElementBuilder
        from youtube_autonomous.elements.builder.image_element_builder import ImageElementBuilder
        from youtube_autonomous.elements.builder.ai_video_element_builder import AIVideoElementBuilder
        from youtube_autonomous.elements.builder.video_element_builder import VideoElementBuilder
        from youtube_autonomous.elements.builder.custom_stock_element_builder import CustomStockElementBuilder
        from youtube_autonomous.elements.builder.stock_element_builder import StockElementBuilder
        from youtube_autonomous.elements.builder.meme_element_builder import MemeElementBuilder
        from youtube_autonomous.elements.builder.sound_element_builder import SoundElementBuilder
        from youtube_autonomous.elements.builder.youtube_video_element_builder import YoutubeVideoElementBuilder
        from youtube_autonomous.elements.builder.text_element_builder import TextElementBuilder
        from youtube_autonomous.elements.builder.premade_element_builder import PremadeElementBuilder
        from youtube_autonomous.elements.builder.effect_element_builder import EffectElementBuilder
        from youtube_autonomous.elements.builder.greenscreen_element_builder import GreenscreenElementBuilder

        return [
            AIImageElementBuilder,
            AIVideoElementBuilder,
            ImageElementBuilder,
            VideoElementBuilder,
            CustomStockElementBuilder,
            StockElementBuilder,
            MemeElementBuilder,
            SoundElementBuilder,
            YoutubeVideoElementBuilder,
            TextElementBuilder,
            PremadeElementBuilder,
            EffectElementBuilder,
            GreenscreenElementBuilder
        ]

    @staticmethod
    def get_subclasses_as_str():
        return ', '.join(ElementBuilder.get_subclasses())
    
    @staticmethod
    def get_subclass_by_type(type: Union[SegmentType, EnhancementType, str]):
        type = ElementParameterValidator.validate_segment_or_enhancement_type(type)

        from youtube_autonomous.elements.builder.ai_image_element_builder import AIImageElementBuilder
        from youtube_autonomous.elements.builder.image_element_builder import ImageElementBuilder
        from youtube_autonomous.elements.builder.ai_video_element_builder import AIVideoElementBuilder
        from youtube_autonomous.elements.builder.video_element_builder import VideoElementBuilder
        from youtube_autonomous.elements.builder.custom_stock_element_builder import CustomStockElementBuilder
        from youtube_autonomous.elements.builder.stock_element_builder import StockElementBuilder
        from youtube_autonomous.elements.builder.meme_element_builder import MemeElementBuilder
        from youtube_autonomous.elements.builder.sound_element_builder import SoundElementBuilder
        from youtube_autonomous.elements.builder.youtube_video_element_builder import YoutubeVideoElementBuilder
        from youtube_autonomous.elements.builder.text_element_builder import TextElementBuilder
        from youtube_autonomous.elements.builder.premade_element_builder import PremadeElementBuilder
        from youtube_autonomous.elements.builder.effect_element_builder import EffectElementBuilder
        from youtube_autonomous.elements.builder.greenscreen_element_builder import GreenscreenElementBuilder

        return {
            SegmentType.MEME: MemeElementBuilder,
            EnhancementType.MEME: MemeElementBuilder,
            SegmentType.AI_IMAGE: AIImageElementBuilder, 
            EnhancementType.AI_IMAGE: AIImageElementBuilder,
            SegmentType.AI_VIDEO: AIVideoElementBuilder,
            EnhancementType.AI_VIDEO: AIVideoElementBuilder,
            SegmentType.IMAGE: ImageElementBuilder,
            EnhancementType.IMAGE: ImageElementBuilder,
            SegmentType.VIDEO: VideoElementBuilder,
            EnhancementType.VIDEO: VideoElementBuilder,
            SegmentType.STOCK: StockElementBuilder,
            EnhancementType.STOCK: StockElementBuilder,
            SegmentType.CUSTOM_STOCK: CustomStockElementBuilder,
            EnhancementType.CUSTOM_STOCK: CustomStockElementBuilder,
            SegmentType.SOUND: SoundElementBuilder,
            EnhancementType.SOUND: SoundElementBuilder,
            SegmentType.YOUTUBE_VIDEO: YoutubeVideoElementBuilder,
            EnhancementType.YOUTUBE_VIDEO: YoutubeVideoElementBuilder,
            SegmentType.TEXT: TextElementBuilder,
            EnhancementType.TEXT: TextElementBuilder,
            SegmentType.PREMADE: PremadeElementBuilder,
            EnhancementType.PREMADE: PremadeElementBuilder,
            EnhancementType.EFFECT: EffectElementBuilder,
            EnhancementType.GREENSCREEN: GreenscreenElementBuilder
        }[type]
    
    @classmethod
    def build_narration(
        cls,
        text: str,
        output_filename: Union[str, None]
    ):
        """
        Generates a narration file that narrates the 'text' provided and
        is stored locally as 'output_filename'. If 'text' or 
        'output_filename' fields are not provided it will raise an 
        Exception.
        """
        ParameterValidator.validate_string_mandatory_parameter('text', text)

        return create_ai_narration(
            text,
            output_filename = Output.get_filename(output_filename, FileTypeX.AUDIO)
        )
    
    @classmethod
    def extract_extra_params(
        cls,
        element,
        method,
        params_to_ignore: list[str] = ['self', 'cls', 'args', 'kwargs']
    ):
        """
        Returns an object containing 'mandatory' and 'optional' fields that
        are the parameters extracted from the provided 'element' 'extra_data'
        field and according to the also provided 'method' (the parameters in
        the signature of that 'method' are the ones that are extracted).

        This method raises an exception if some of the 'mandatory' fields are
        not given in the 'extra_data' field.
        """
        parameters_needed = ParameterObtainer.get_parameters_from_method(method, params_to_ignore)  
        parameters = {}

        extra_params = (
            element.get('extra_params')
            if PythonValidator.is_dict(element) else
            element.extra_params
        )

        if (
            not extra_params and
            not PythonValidator.is_dict(extra_params)
        ):
            raise Exception('No "extra_params" found.')

        missing_mandatory_parameters = []
        for mandatory_parameter in parameters_needed.mandatory_parameters:
            if not mandatory_parameter in extra_params:
                missing_mandatory_parameters.append(mandatory_parameter)
            else:
                parameters[mandatory_parameter] = extra_params[mandatory_parameter]

        if len(missing_mandatory_parameters) > 0:
            missing_params_str = ', '.join(missing_mandatory_parameters)
            raise Exception(f'The extra parameters "{missing_params_str}" are missing on the provided element and they are needed to build it.')

        parameters.update({
            key: extra_params[key]
            for key in parameters_needed.optional_parameters
            if key in extra_params
        })

        return parameters
    
    @classmethod
    def handle_narration_from_segment(cls, segment: dict):
        pass